"""esbmtk: A general purpose Earth Science box model toolkit Copyright
     (C), 2020-2021 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
     published by the Free Software Foundation, either version 3 of
     the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see
     <https://www.gnu.org/licenses/>.

"""
from __future__ import annotations
import typing as tp
import numpy as np
from math import log, sqrt
from esbmtk import ReservoirGroup

# if tp.TYPE_CHECKING:
#     from .esbmtk import ReservoirGroup, Flux


def get_hplus(
    rg: ReservoirGroup,
    dic: float,
    ta: float,
    hplus: float,
) -> float:

    k1 = rg.swc.K1  # K1
    k2 = rg.swc.K2  # K2
    KW = rg.swc.KW  # KW
    KB = rg.swc.KB  # KB
    boron = rg.swc.boron  # boron
    hplus_0 = hplus

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1 * k1 - 4 * k1 * k2 * (1 - (2 * gamm))
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))

    return hplus - hplus_0


def carbonate_system_1_ode(
    swc: any,
    dic: float,
    ta: float,
    hplus: float,
) -> float:

    """Calculates and returns the H+ and carbonate alkalinity concentrations
     for the given reservoirgroup

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    """

    k1 = swc.K1  # K1
    k1k1 = swc.K1K1  #  K1 * K1
    k1k2 = swc.K1K2  # K1 * K2
    KW = swc.KW  # KW
    KB = swc.KB  # KB
    boron = swc.boron  # boron
    hplus_0 = hplus

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1k1 - k1k2 * (4 - 8 * gamm)
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))
    co2aq: float = dic / (1 + (k1 / hplus) + (k1k2 / (hplus * hplus)))
    diff = hplus - hplus_0

    return diff, co2aq


# @njit(parallel=False, fastmath=True, error_model="numpy")
def carbonate_system_2_ode(
    rg: ReservoirGroup,  # 2 Reservoir handle
    Bm: float,  # 3 CaCO3 export flux as DIC
    dic_db: float,  # 4 DIC in the deep box
    ta_db: float,  # 5 TA in the deep box
    dic_sb: float,  # 6 [DIC] in the surface box
    dic_sb_l: float,  # 7 [DIC_l] in the surface box
    hplus_0: float,  # 8 hplus in the deep box at t-1
    zsnow: float,  # 9 snowline in meters below sealevel at t-1
) -> tuple:
    """Calculates and returns the fraction of the carbonate rain that is
    dissolved an returned back into the ocean. This functions returns:

    DIC_burial, DIC_burial_l, Hplus, zsnow

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004
    """

    # Parameters
    k1 = rg.swc.K1  # K1
    k1k1 = rg.swc.K1K1
    k2 = rg.swc.K2  # K2
    k1k2 = rg.swc.K1K2  # K2
    KW = rg.swc.KW  # KW
    KB = rg.swc.KB  # KB
    boron = rg.swc.boron  # boron
    p = rg.cs.function_params
    ksp0 = p[5]
    ca2 = rg.swc.ca2  # Ca2+
    kc = p[6]
    AD = p[8]
    zsat0 = int(abs(p[9]))
    I_caco3 = p[12]
    alpha = p[13]
    zsat_min = int(abs(p[14]))
    zmax = int(abs(p[15]))
    z0 = int(abs(p[16]))
    depth_area_table = rg.cs.depth_area_table
    area_dz_table = rg.cs.area_dz_table
    Csat_table = rg.cs.Csat_table

    # calc carbonate alkalinity based t-1
    oh: float = KW / hplus_0
    boh4: float = boron * KB / (hplus_0 + KB)
    fg: float = hplus_0 - oh - boh4
    ca: float = ta_db + fg

    # calculate carbon speciation
    # The following equations are after Follows et al. 2006
    gamm: float = dic_db / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1k1 - k1k2 * (4 - 8 * gamm)
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))
    co3 = max(dic_db / (1 + hplus / k2 + hplus * hplus / k1k2), 3.7e-05)
    # ---------- compute critical depth intervals eq after  Boudreau (2010)
    # all depths will be positive to facilitate the use of lookup_tables
    zsat = int(zsat0 * log(ca2 * co3 / ksp0))
    zsat = np.clip(zsat, zsat_min, zmax)
    zcc = int(zsat0 * log(Bm * ca2 / (ksp0 * AD * kc) + ca2 * co3 / ksp0))  # eq3
    zcc = np.clip(zcc, zsat_min, zmax)
    # get fractional areas
    B_AD = Bm / AD
    A_z0_zsat = depth_area_table[z0] - depth_area_table[zsat]
    A_zsat_zcc = depth_area_table[zsat] - depth_area_table[zcc]
    A_zcc_zmax = depth_area_table[zcc] - depth_area_table[zmax]
    # ------------------------Calculate Burial Fluxes----------------------------- #
    BCC = A_zcc_zmax * B_AD
    BNS = alpha * A_z0_zsat * B_AD
    diff_co3 = Csat_table[zsat:zcc] - co3
    area_p = area_dz_table[zsat:zcc]
    BDS_under = kc * area_p.dot(diff_co3)
    BDS_resp = alpha * (A_zsat_zcc * B_AD - BDS_under)
    BDS = BDS_under + BDS_resp
    diff: np.ndarray = Csat_table[zcc : int(zsnow)] - co3
    area_p: np.ndarray = area_dz_table[zcc : int(zsnow)]
    # integrate saturation difference over area
    BPDC = kc * area_p.dot(diff)
    BPDC = max(BPDC, 0)  # prevent negative values
    d_zsnow = -BPDC / (area_dz_table[int(zsnow)] * I_caco3)
    # get isotope ratio in reservoir
    BD: float = BDS + BCC + BNS + BPDC
    """Bm is the flux of CaCO3 into the box. However, the model should
    use the bypass option and leave all flux calculations to the
    cs_code.  As such, we simply add the fraction of the input flux
    that dissolves, and ignore the fraction that is buried.  

    The isotope ratio of the dissolutio flux is determined by the delta
    value of the sediments we are dissolving, and the delta of the carbonate rain.
    The currrent code, assumes that both are the same.
    """
    BD_l = BD * dic_sb_l / dic_sb
    dH = hplus - hplus_0

    return -BD, -BD_l, dH, d_zsnow


def gas_exchange_ode(scale, gas_c, p_H2O, solubility, g_c_aq) -> float:
    """Calculate the gas exchange flux across the air sea interface

    Parameters:
    scale: surface area in m^2
    gas_c: species concentration in atmosphere
    p_H2O: water vapor partial pressure
    solubility: species solubility  mol/(m^3 atm)
    gc_aq: concentration of the dissolved gas in water
    """
    beta = solubility * (1 - p_H2O)
    f = scale * (gas_c * beta - g_c_aq * 1e3)

    return -f


def gas_exchange_ode_with_isotopes(
    scale,  # surface area in m^2 * piston velocity
    gas_c,  # species concentration in atmosphere
    gas_c_l,  # same but for the light isotope
    liquid_c,  # c of the reference species (e.g., DIC)
    liquid_c_l,  # same but for the light isotopeof DIC
    p_H2O,  # water vapor pressure
    solubility,  # solubility constant
    gas_c_aq,  # Gas concentration in liquid phase
    a_db,  # fractionation factor between dissolved CO2aq and HCO3-
    a_dg,  # fractionation between CO2aq and CO2g
    a_u,  # kinetic fractionation during gas exchange
) -> tuple(float, float):
    """Calculate the gas exchange flux across the air sea interface
    for co2 including isotope effects.

    Note that the sink delta is co2aq as returned by the carbonate VR
    this equation is for mmol but esbmtk uses mol, so we need to
    multiply by 1E3

    The Total flux across interface dpends on the difference in either
    concentration or pressure the atmospheric pressure is known, as gas_c, and
    we can calculate the equilibrium pressure that corresponds to the dissolved
    gas in the water as [CO2]aq/beta.

    Conversely, we can convert the the pCO2 into the amount of dissolved CO2 =
    pCO2 * beta

    The h/c ratio in HCO3 estimated via h/c in DIC. Zeebe writes C12/C13 ratio
    but that does not work. the C13/C ratio results however in -8 permil
    offset, which is closer to observations
    """
    # Solibility with correction for pH2O
    beta = solubility * (1 - p_H2O)
    # f as afunction of solubility difference
    f = scale * (beta * gas_c - gas_c_aq * 1e3)
    # isotope ratio of DIC
    Rt = (liquid_c - liquid_c_l) / liquid_c
    # get heavy isotope concentrations in atmosphere
    gas_c_h = gas_c - gas_c_l  # gas heavy isotope concentration
    # get exchange of the heavy isotope
    f_h = scale * a_u * (a_dg * gas_c_h * beta - Rt * a_db * gas_c_aq * 1e3)
    f_l = f - f_h  # the corresponding flux of the light isotope

    return -f, -f_l



import pytest


@pytest.fixture(params=[True, False])
def create_model(request):

    # from module import symbol
    from esbmtk import Model, Reservoir

    c0 = 3.1
    d0 = 0
    v0 = 1025
    # create model
    M1 = Model(
        name="M1",  # model name
        stop="1 kyrs",  # end time of model
        timestep=" 1 yr",  # base unit for time
        mass_unit="mol",  # base unit for mass
        volume_unit="l",  # base unit for volume
        concentration_unit="mol/kg",
        element="Carbon",  # load default element and species definitions
        m_type="both",
        save_flux_data=request.param,
        ideal_water=True,
    )
    Reservoir(
        name="R1",  # Name of reservoir
        species=M1.Carbon.DIC,  # Species handle
        delta=d0,  # initial delta
        concentration=f"{c0} mol/kg",  # concentration
        volume=f"{v0} l",  # reservoir size (m^3)
        register=M1,
    )

    return v0, c0, d0, M1

@pytest.fixture(params=[True, False])
def create_model_with_seawater(request):

    # from module import symbol
    from esbmtk import Model, Reservoir, Q_

    c0 = 3.1
    d0 = 0
    v0 = 1025
    # create model
    M1 = Model(
        name="M1",  # model name
        stop="1 kyrs",  # end time of model
        timestep=" 1 yr",  # base unit for time
        mass_unit="mol",  # base unit for mass
        volume_unit="l",  # base unit for volume
        concentration_unit="mol/kg",
        element="Carbon",  # load default element and species definitions
        m_type="both",
        save_flux_data=request.param,
        ideal_water=False,
    )
    Reservoir(
        name="R1",  # Name of reservoir
        species=M1.Carbon.DIC,  # Species handle
        delta=d0,  # initial delta
        concentration=f"{c0} mol/kg",  # concentration
        volume=f"{v0} l",  # reservoir size (m^3)
        register=M1,
    )

    return v0, c0, d0, M1

def test_delta_conversion():
    from esbmtk import get_imass, get_delta, get_flux_data

    r = 0.0112372
    m = 1
    d = 0

    l, h = get_imass(m, d, r)
    d2 = get_delta(l, h, r)
    assert abs(d - d2) < 1e-10, "deltas should be similar"
    A = get_flux_data(m, d, r)

    assert A[0] == m, "mass should be equal"
    assert A[1] == l, "light mass should be equal"
    assert A[2] == h, "heavy mass should be equal"
    assert A[3] == d, "deltas should be equal"


def test_delta_conversion_with_vector():
    from esbmtk import get_imass, get_delta, get_flux_data
    import numpy as np

    r = 0.0112372
    m = np.zeros(4) + 3001
    d = 0
    l, h = get_imass(m, d, r)
    d2 = get_delta(l, h, r)
    assert abs(max(d2 - d)) < 1e-10


def test_isotope_mass_calculation():
    from esbmtk import get_imass

    r = 0.0112372

    d = 21
    m = 10

    li, hi = get_imass(m, d, r)

    assert m - li - hi == 0


def test_fractionation_calculation():
    from esbmtk import get_imass, get_delta, get_frac

    r = 0.044162589
    m = 10
    d = 21
    a = 1.003

    l, h = get_imass(m, d, r)
    l2, h2 = get_frac(m, l, a)

    a2 = (l * h2) / (h * l2)

    diff = abs(a - a2)
    assert diff < 1e-15


def test_reservoir_creation(create_model):
    """Test that reservoir is initialized correctly"""
    v0, c0, d0, M1 = create_model

    mass = v0 * c0
    assert M1.R1.c[2] == c0
    assert M1.R1.m[2] == mass
    M1.get_delta_values()
    assert abs(M1.R1.d[2] - d0) < 1e-10


@pytest.mark.parametrize(
    "idx", [[0, 200], [100, 200], [200, 400], [200, 1200], [1100, 1000]]
)
@pytest.mark.parametrize("stype", ["square", "pyramid", "bell"])
def test_signal_indexing(idx, stype):
    """
    test that signal indexing works
    """

    from esbmtk import Source, Sink, Reservoir, Connect, Signal, Model
    import numpy as np

    c0 = 3.1
    d0 = 0
    v0 = 1025
    # create model
    M1 = Model(
        name="M1",  # model name
        stop="1 kyrs",  # end time of model
        timestep=" 1 yr",  # base unit for time
        mass_unit="mol",  # base unit for mass
        volume_unit="l",  # base unit for volume
        element="Carbon",  # load default element and species definitions
        concentration_unit="mol/kg",
        m_type="both",
        offset="100 yrs",
    )

    Source(name="SO1", species=M1.Carbon.DIC, register=M1)

    Sink(name="SI1", species=M1.Carbon.DIC, register=M1)
    Sink(name="SI2", species=M1.Carbon.DIC, register=M1)

    Reservoir(
        name="R1",  # Name of reservoir
        species=M1.Carbon.DIC,  # Species handle
        delta=d0,  # initial delta
        concentration=f"{c0} mol/kg",  # concentration
        volume=f"{v0} l",  # reservoir size (m^3)
        register=M1,
    )

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Signal(
        register=M1,
        name="foo",
        species=M1.Carbon.DIC,
        start=f"{idx[0]} yrs",
        duration=f"{idx[1]} yrs",
        shape=stype,
        magnitude="50 mol/year",
        delta=-28,
    )

    Signal(
        register=M1,
        name="foo2",
        species=M1.Carbon.DIC,
        start=f"{idx[0]} yrs",
        duration=f"{idx[1]} yrs",
        shape=stype,
        mass="5000 mol",
        delta=-28,
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        delta=0,
        signal=M1.foo,
        # alpha=-28,  # set a default flux
    )

    M1.run()


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_fractionation(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation

    """

    from esbmtk import Source, Sink, Connect, Reservoir

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="100 mol/yr",  # weathering flux in
        alpha=-28,  # set a default flux
    )
    M1.run(solver=solver)
    M1.get_delta_values()
    assert round(M1.R1.d[-2], 4) == 28.8066


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_scale_flux(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        alpha=-28,  # set a default flux
    )

    names = M1.flux_summary(filter_by="default_b", return_list=True)
    Connect(
        id="default_c",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI2,  # target of flux
        ctype="scale_with_flux",
        ref_flux=names[0],
        scale=1,  # weathering flux in
    )
    M1.run(solver=solver)
    M1.get_delta_values()
    assert round(M1.R1.d[-2], 2) == 14.2


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_scale_with_concentration_empty(create_model, solver):

    from esbmtk import Source, Sink, Connect, Reservoir
    import numpy as np

    v0, c0, d0, M1 = create_model
    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Reservoir(
        name="R2",  # Name of reservoir
        species=M1.Carbon.DIC,  # Species handle
        delta=d0,  # initial delta
        concentration=f"0 mol/kg",  # concentration
        volume=f"{v0} l",  # reservoir size (m^3)
        register=M1,
    )
    Connect(
        id="default_a",
        register=M1,
        source=M1.R2,  # source of flux
        sink=M1.R1,  # target of flux
        ctype="scale_with_concentration",
        alpha=-70,
    )
    M1.run(solver=solver)
    M1.get_delta_values()

    names = M1.flux_summary(filter_by="default_a", return_list=True)
    assert M1.R2.m[500] == 0
    assert M1.R2.l[500] == 0
    assert M1.R2.c[500] == 0
    assert M1.R1.c[500] == 3.1
    assert round(M1.R1.d[500], 10) == 0
    f, fl = names[0].fa
    assert f == 0
    assert fl == 0


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_scale_with_concentration(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        alpha=-28,  # set a default flux
    )

    Connect(
        id="default_c",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI2,  # target of flux
        ctype="scale_with_concentration",
        scale=1,
    )
    M1.run(solver=solver)
    M1.get_delta_values()

    assert round(M1.R1.d[-2]) == 14
    assert round(M1.R1.c[-2]) == 32


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_scale_with_mass(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        alpha=-28,  # set a default flux
    )

    Connect(
        id="default_c",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI2,  # target of flux
        ctype="scale_with_mass",
        scale=1 / 1025,
    )
    M1.run(solver=solver)
    M1.get_delta_values()

    assert round(M1.R1.d[-2]) == 14
    assert round(M1.R1.c[-2]) == 32


# should add tests for magnitude vs mass
@pytest.mark.parametrize("solver", ["numba", "python"])
def test_square_signal(create_model, solver):
    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    import numpy as np

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Signal(
        register=M1,
        name="foo",
        species=M1.Carbon.DIC,
        start="100 yrs",
        duration="800 yrs",
        shape="square",
        magnitude="50 mol/year",
        delta=-28,
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        delta=0,
        signal=M1.foo,
        # alpha=-28,  # set a default flux
    )

    M1.run(solver=solver)
    M1.get_delta_values()

    assert round(M1.R1.d[90], 0) == 0
    assert round(max(M1.R1.d)) == 277
    assert round(M1.R1.c[200]) == 8
    assert round(max(M1.R1.c)) == 13
    # M1.plot([M1.R1, M1.foo, M1.C_R1_2_SI1.R1_2_SI1_F])


# should add tests for magnitude vs mass
@pytest.mark.parametrize("solver", ["numba", "python"])
def test_pyramid_signal(create_model, solver):
    v0, c0, d0, M1 = create_model

    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    import numpy as np

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Signal(
        name="foo",
        species=M1.Carbon.DIC,
        start="100 yrs",
        duration="800 yrs",
        shape="pyramid",
        magnitude="50 mol/year",
        delta=-28,
        register=M1,
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        delta=0,
        signal=M1.foo,
        # alpha=-28,  # set a default flux
    )
    M1.run(solver=solver)
    M1.get_delta_values()

    assert round(max(M1.R1.d)) == 41
    assert round(max(M1.R1.c)) == 32
    assert np.argmax(M1.R1.d) == 702
    # M1.plot([M1.R1, M1.foo, M1.C_R1_2_SI1.R1_2_SI1_F])


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_pyramid_signal_multiplication(create_model, solver):
    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    import numpy as np

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Signal(
        name="foo",
        species=M1.Carbon.DIC,
        start="100 yrs",
        duration="800 yrs",
        shape="pyramid",
        magnitude="1 mol/year",
        stype="multiplication",
        register=M1,
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        delta=10,
        signal=M1.foo,
        # alpha=-28,  # set a default flux
    )

    M1.run(solver=solver)
    M1.get_delta_values()

    assert round(M1.R1.c[-2]) == 81
    assert np.argmin(M1.R1.d) == 708
    assert round(M1.R1.d[-2], 2) == -2.41
    # M1.plot([M1.R1, M1.foo, M1.C_R1_2_SI1.R1_2_SI1_F])


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_pyramid_signal_multiply(create_model, solver):
    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    import numpy as np

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI2", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Signal(
        name="foo",
        species=M1.Carbon.DIC,
        start="100 yrs",
        duration="800 yrs",
        shape="pyramid",
        magnitude="1 mol/year",
        stype="multiplication",
        register=M1,
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="50 mol/yr",  # weathering flux in
        alpha=-70,
        signal=M1.foo,
        # alpha=-28,  # set a default flux
    )

    M1.run(solver=solver)
    M1.get_delta_values()
    # diff = M1.C_R1_2_SI1.R1_2_SI1_F.d[500] - M1.R1.d[500]
    assert round(M1.R1.c[-2]) == 81
    assert np.argmax(M1.R1.d) == 692
    # assert round(diff) == -71
    # M1.plot([M1.R1, M1.foo, M1.C_R1_2_SI1.R1_2_SI1_F])


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_scale_with_flux_and_signal_multiplication(create_model, solver):

    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    import numpy as np

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )
    Signal(
        register=M1,
        name="foo",
        species=M1.Carbon.DIC,
        start="100 yrs",
        duration="800 yrs",
        shape="pyramid",
        magnitude="1 mol/year",
        stype="multiplication",
    )

    names = M1.flux_summary(filter_by="default_a", return_list=True)
    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="scale_with_flux",
        ref_flux=names[0],
        alpha=-72,
        signal=M1.foo,
        # alpha=-28,  # set a default flux
    )
    M1.run(solver=solver)
    M1.get_delta_values()

    # m, l  =  M1.C_R1_2_SI1.R1_2_SI1_F.fa
    # r =  M1.R1.sp.r
    # d = get_delta_h(l, m-l, r)
    # diff = d - M1.R1.d[-2]
    # assert round(diff) == -74
    assert round(M1.R1.c[-2]) == 61
    assert np.argmax(M1.R1.d) == 672
    assert round(max(M1.R1.d)) == 43

    # M1.plot([M1.R1, M1.foo, M1.C_R1_2_SI1.R1_2_SI1_F])


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_scale_with_flux_and_signal_addition(create_model, solver):

    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    import numpy as np

    v0, c0, d0, M1 = create_model

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.R1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )
    Signal(
        register=M1,
        name="foo",
        species=M1.Carbon.DIC,
        start="100 yrs",
        duration="800 yrs",
        shape="pyramid",
        magnitude="100 mol/year",
        stype="addition",
    )

    names = M1.flux_summary(filter_by="default_a", return_list=True)
    Connect(
        id="default_b",
        register=M1,
        source=M1.R1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="scale_with_flux",
        ref_flux=names[0],
        alpha=-72,
        signal=M1.foo,
        scale=0.1
        # alpha=-28,  # set a default flux
    )
    # M1.run(solver='python')
    M1.run(solver=solver)
    M1.get_delta_values()

    # diff = M1.C_R1_2_SI1.R1_2_SI1_F.d[500] - M1.R1.d[500]
    # assert round(diff) == -75
    assert round(M1.R1.c[-2]) == 52
    assert np.argmax(M1.R1.d) == 658
    assert round(max(M1.R1.d)) == 53

    # M1.plot([M1.R1, M1.foo,
    #          M1.C_SO1_2_R1.SO1_2_R1_F,
    #          M1.C_R1_2_SI1.R1_2_SI1_F])


def test_seawaterconstants(create_model_with_seawater):
    from esbmtk import Source, Sink, Reservoir, Connect, Signal
    from esbmtk import ExternalCode, ReservoirGroup, add_carbonate_system_1
    from esbmtk import GasReservoir, AirSeaExchange
    import numpy as np
    import sys

    v0, c0, d0, M1 = create_model_with_seawater
    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    ReservoirGroup(
        name="S1",  # Name of reservoir group
        # volume = "1E5 l",       # see below
        geometry=[0, 6000, 1],
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.1 mmol/kg", M1.Carbon.TA: "2.36 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        seawater_parameters={"temperature": 5, "pressure": 0, "salinity": 0},
        # @carbonate_system= True,
        register=M1,
    )
    assert round(M1.S1.swc.density, 6) == 999.966751

    ReservoirGroup(
        name="S2",  # Name of reservoir group
        # volume = "1E5 l",       # see below
        geometry=[0, 6000, 1],
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.1 mmol/kg", M1.Carbon.TA: "2.36 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        seawater_parameters={"temperature": 25, "pressure": 1000, "salinity": 35},
        # @carbonate_system= True,
        register=M1,
    )

    assert round(M1.S2.swc.density, 6) == 1062.538172  # kg/m^3
    assert round(M1.S2.swc.so4, 6) == 0.028253  # mol/kgiter


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_carbonate_system1_constants(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir

    v0, c0, d0, M1 = create_model

    from esbmtk import ExternalCode, ReservoirGroup, add_carbonate_system_1
    import numpy as np

    ReservoirGroup(
        name="S",  # Name of reservoir group
        volume="1E5 l",  # see below
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.1 mmol/kg", M1.Carbon.TA: "2.4 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        seawater_parameters={"temperature": 25, "pressure": 1, "salinity": 35},
        # @carbonate_system= True,
        register=M1,
    )

    add_carbonate_system_1([M1.S])

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    # DIC influx
    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.S.DIC,  # target of flux
        rate="100 mol/yr",  # weathering flux in
    )
    # TA influx
    Connect(
        id="default_b",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.S.TA,  # target of flux
        rate="100 mol/yr",  # weathering flux in
    )

    # DIC outflux
    Connect(
        id="default_c",
        register=M1,
        source=M1.S.DIC,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="100 mol/yr",  # weathering flux in
    )
    # TA out
    Connect(
        id="default_d",
        register=M1,
        source=M1.S.TA,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="100 mol/yr",  # weathering flux in
        # alpha=-28,  # set a default flux
    )

    M1.run(solver=solver)
    i = 200
    assert round(M1.S.DIC.c[i] * 1000, 2) == 2.1  # DIC
    assert round(M1.S.TA.c[i] * 1000, 2) == 2.4  # TA
    assert round(-np.log10(M1.S.cs.H[i]), 2) == 8  # pH
    assert round(M1.S.cs.HCO3[i] * 1000, 2) == 1.86
    assert round(M1.S.cs.CO3[i] * 1000, 2) == 0.22
    assert round(M1.S.cs.CO2aq[i] * 1000, 4) == 0.0133


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_carbonate_system2_(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir

    v0, c0, d0, M1 = create_model

    from esbmtk import ExternalCode, ReservoirGroup, add_carbonate_system_2
    import numpy as np

    ReservoirGroup(
        name="S",  # Name of reservoir group
        # volume = "1E5 l",       # see below
        geometry=[0, 6000, 1],
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.3 mmol/kg", M1.Carbon.TA: "2.4 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        seawater_parameters={"temperature": 2, "pressure": 240, "salinity": 35},
        # @carbonate_system= True,
        register=M1,
    )

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    # DIC influx
    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.S.DIC,  # target of flux
        rate="100 mol/yr",  # weathering flux in
    )
    # TA influx
    Connect(
        id="default_b",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.S.TA,  # target of flux
        rate="100 mol/yr",  # weathering flux in
    )

    # DIC outflux
    Connect(
        id="default_c",
        register=M1,
        source=M1.S.DIC,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="100 mol/yr",  # weathering flux in
        bypass="sink",
    )
    # TA x
    Connect(
        id="default_d",
        register=M1,
        source=M1.S.TA,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="100 mol/yr",  # weathering flux in
        bypass="sink",
        # alpha=-28,  # set a default flux
    )
    fs = M1.flux_summary(filter_by="default_a", return_list=True)
   
    exp = fs[0]
    exp.fa = np.array([60e12, 60e12])
    add_carbonate_system_2(
        r_db=[M1.S],
        r_sb=[M1.S],
        carbonate_export_fluxes=[exp],
        alpha=0.6,
        zsat_min=-200,
        z0=-200,
    )
    M1.run(solver=solver)
    i = 997

    """Below numbers need to be verified
    """
    assert round(M1.S.DIC.c[i] * 1000, 2) == 2.39  # DIC
    assert round(M1.S.TA.c[i] * 1000, 2) == 2.49  # TA
    assert round(-np.log10(M1.S.cs.H[i]), 2) == 7.84  # pH
    assert round(M1.S.cs.HCO3[i] * 1000, 2) == 2.27
    # assert round(M1.S.cs.CO3[i] * 1000, 4) == 0.0661
    # assert round(M1.S.cs.CO2aq[i] * 1000, 4) == 0.0386
    # assert round(M1.S.cs.zsat[i]) == 2335
    # assert round(M1.S.cs.zsnow[i]) == 4605


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_gas_exchange(create_model_with_seawater, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir, GasReservoir
    from esbmtk import ExternalCode, ReservoirGroup, AirSeaExchange
    from esbmtk import add_carbonate_system_1
    import numpy as np

    v0, c0, d0, M1 = create_model_with_seawater

    Source(register=M1, name="SO1", species=M1.Carbon.DIC)
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    ReservoirGroup(
        name="S",  # Name of reservoir group
        # volume = "1E5 l",       # see below
        geometry=[0, 6000, 1],
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.1 mmol/kg", M1.Carbon.TA: "2.36 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        seawater_parameters={"temperature": 25, "pressure": 1, "salinity": 35},
        # @carbonate_system= True,
        register=M1,
    )

    add_carbonate_system_1(rgs=[M1.S])

    GasReservoir(
        name="CO2_At",
        species=M1.Carbon.CO2,
        reservoir_mass="1.833E20 mol",
        species_ppm="280 ppm",
        isotopes=True,
        delta=-7,
        register=M1,
    )

    # DIC influx
    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.S.DIC,  # target of flux
        rate="0 mol/yr",  # weathering flux in
        delta=0,
    )

    AirSeaExchange(
        gas_reservoir=M1.CO2_At,  # Reservoir
        liquid_reservoir=M1.S.DIC,  # ReservoirGroup
        species=M1.Carbon.CO2,
        ref_species=M1.S.cs.CO2aq,
        solubility=M1.S.swc.SA_co2,  # float
        area=M1.S.area,
        piston_velocity="4.8 m/d",
        water_vapor_pressure=M1.S.swc.p_H2O,
        id="A_sb",
        register=M1,
    )

    M1.run(solver=solver)

    M1.get_delta_values()
    # calculate theoretical equlibrium pCO2 based on pCO2
    epco2_at = M1.CO2_At.c * (1 - M1.S.swc.p_H2O) * 1e6
    # calculate equilibrium pCO2 based on CO2aq
    epco2_aq = M1.S.cs.CO2aq / M1.S.swc.K0 * 1e6

    diff_c = epco2_at[-3] - epco2_aq[-3]
    diff_d = M1.S.DIC.d[-3] - M1.CO2_At.d[-3]

    assert round(abs(diff_c), 1) == 9.8  # ppm
    assert round(diff_d, 1) == 8.0


@pytest.mark.parametrize(
    "solver, scale",
    [
        ("numba", 1),
        ("numba", 0.5),
        ("python", 1),
        ("python", 0.5),
    ],
)
def test_weathering(create_model, solver, scale):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir, GasReservoir
    from esbmtk import ExternalCode, ReservoirGroup, AirSeaExchange
    from esbmtk import get_delta_i, Q_
    import numpy as np

    v0, c0, d0, M1 = create_model

    ReservoirGroup(
        name="S",  # Name of reservoir group
        # volume = "1E5 l",       # see below
        geometry=[0, 6000, 1],
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.3 mmol/kg", M1.Carbon.TA: "2.4 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        register=M1,
    )

    Source(register=M1, name="Fw", species=M1.Carbon.DIC)

    GasReservoir(
        name="CO2_At",
        species=M1.Carbon.CO2,
        reservoir_mass="1.833E20 mol",
        species_ppm="400 ppm",
        isotopes=True,
        delta=0,
        register=M1,
    )

    Connect(
        register=M1,
        source=M1.Fw,  # source of flux
        sink=M1.S.DIC,
        reservoir_ref=M1.CO2_At,
        ctype="weathering",
        id="we",
        scale=scale,
        ex=0.4,
        pco2_0="280 ppm",
        rate="1E13 mol/yr",
        delta=5,
    )

    M1.run(solver=solver)
    names = M1.flux_summary(filter_by="we", return_list=True)
    f, fl = names[0].fa
    # flux computed at 400ppm
    tf = scale * 1e13 * (400 / 280) ** 0.4
    # f = M1.C_Fw_2_DIC_we.Fw_2_DIC_we_F.fa[0]
    # test that flux calculation is correct
    assert tf - f == 0
    # fl = M1.C_Fw_2_DIC_we.Fw_2_DIC_we_F.fa[1] 
    d = get_delta_i(fl, f - fl, M1.Carbon.DIC.r)
    # test that isotope calculation is correct
    assert d - 5 < 0.000001


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_weathering_with_atmosphere_as_source(create_model, solver):
    """Test that isotope fractionation from a flux out of a reservoir
    results in the corrrect fractionation, when using the numba solver

    """

    from esbmtk import Source, Sink, Connect, Reservoir, GasReservoir
    from esbmtk import ExternalCode, ReservoirGroup, AirSeaExchange
    from esbmtk import get_delta_i, Q_
    import numpy as np

    v0, c0, d0, M1 = create_model

    ReservoirGroup(
        name="S",  # Name of reservoir group
        # volume = "1E5 l",       # see below
        geometry=[0, 6000, 1],
        delta={M1.Carbon.DIC: 0, M1.Carbon.TA: 0},  # dict of delta values
        concentration={M1.Carbon.DIC: "2.3 mmol/kg", M1.Carbon.TA: "2.4 mmol/kg"},
        isotopes={M1.Carbon.DIC: True},
        register=M1,
    )

    Source(register=M1, name="Fw", species=M1.Carbon.DIC)

    GasReservoir(
        name="CO2_At",
        species=M1.Carbon.CO2,
        reservoir_mass="1.833E20 mol",
        species_ppm="400 ppm",
        isotopes=True,
        delta=12,
        register=M1,
    )

    Connect(
        register=M1,
        source=M1.CO2_At,  # source of flux
        sink=M1.S.DIC,
        reservoir_ref=M1.CO2_At,
        ctype="weathering",
        id="we",
        scale=1,
        ex=0.4,
        pco2_0="280 ppm",
        rate="1 mol/yr",
        delta=5,
    )

    M1.run(solver=solver)
    names = M1.flux_summary(filter_by="we", return_list=True)
    f, fl = names[0].fa
    # f = M1.C_CO2_At_2_DIC_we.CO2_At_2_DIC_we_F.fa[0]
    # fl = M1.C_CO2_At_2_DIC_we.CO2_At_2_DIC_we_F.fa[1]
    d = get_delta_i(fl, f - fl, M1.Carbon.DIC.r)

    tf = 1 * (400 / 280) ** 0.4
    assert tf - f == 0
    assert d - 12 < 0.00001


@pytest.mark.parametrize("solver", ["numba", "python"])
def test_gasreservoir_flux_alpha(create_model, solver):
    from esbmtk import Source, Sink, Reservoir, Connect, Signal, GasReservoir
    import numpy as np

    v0, c0, d0, M1 = create_model
    Source(register=M1, name="SO1", species=M1.Carbon.DIC)

    GasReservoir(
        name="GR1",
        species=M1.Carbon.CO2,
        delta=d0,  # initial delta
        reservoir_mass="1025E4 mol",
        species_ppm=f"300 ppm",
        isotopes=True,
        register=M1,
    )
    Sink(register=M1, name="SI1", species=M1.Carbon.DIC)

    Connect(
        id="default_a",
        register=M1,
        source=M1.SO1,  # source of flux
        sink=M1.GR1,  # target of flux
        rate="100 mol/yr",  # weathering flux in
        delta=d0,  # set a default flux
    )

    Connect(
        id="default_b",
        register=M1,
        source=M1.GR1,  # source of flux
        sink=M1.SI1,  # target of flux
        ctype="Regular",
        rate="100 mol/yr",  # weathering flux in
        alpha=-28,  # set a default flux
    )

    M1.run(solver=solver)
    M1.get_delta_values()
    assert round(M1.GR1.d[-2], 4) == 28.8066


# the following do currently not work with numba

# @pytest.mark.parametrize("solver", ["numba","python"])
# def test_sum_fluxes(create_model,solver):
#     """
#     test that the code which adds fluxes to a reservoir yields the expected results
#     """

#     from esbmtk import Connect, Source, Sink

#     v0, c0, d0, M1 = create_model

#     Source(name="SO1", species=M1.Carbon.CO2)
#     Source(name="SO2", species=M1.Carbon.CO2)
#     Sink(name="SI1", species=M1.Carbon.CO2)
#     Sink(name="SI2", species=M1.Carbon.CO2)

#     Connect(
#         source=M1.SO1,  # source of flux
#         sink=M1.R1,  # target of flux
#         rate="100 mol/yr",  # weathering flux in
#         delta=d0,  # set a default flux
#     )
#     Connect(
#         source=M1.SO2,  # source of flux
#         sink=M1.R1,  # target of flux
#         rate="300 mol/yr",  # weathering flux in
#         delta=d0,  # set a default flux
#     )

#     Connect(
#         source=M1.R1,  # source of flux
#         sink=M1.SI1,  # target of flux
#         rate="250 mol/yr",  # weathering flux in
#         delta=d0,  # set a default flux
#     )

#     Connect(
#         source=M1.R1,  # source of flux
#         sink=M1.SI2,  # target of flux
#         rate="150 mol/yr",  # weathering flux in
#         delta=d0,  # set a default flux
#     )

#     M1.run(solver=solver)
# M1.get_delta_values()

#     assert M1.R1.c[-2] == c0
#     M1.R1.d[-2] == d0

#     # strangely, this fails at the moment
#     # assert R1.d[-]2 == 1

# @pytest.mark.parametrize("solver", ["numba","python"])
# def test_passive_sum(create_model,solver):
#     """
#     test that the code which adds fluxes to a reservoir yields the expected results
#     """

#     from esbmtk import Connect, Source, Sink

#     v0, c0, d0, M1 = create_model

#     Source(name="SO1", species=M1.Carbon.CO2)
#     Source(name="SO2", species=M1.Carbon.CO2)
#     Sink(name="SI1", species=M1.Carbon.CO2)
#     Sink(name="SI2", species=M1.Carbon.CO2)

#     Connect(
#         source=M1.SO1,  # source of flux
#         sink=M1.R1,  # target of flux
#         rate="100 mol/yr",  # weathering flux in
#         delta=d0,  # set a default flux
#     )
#     Connect(
#         source=M1.SO2,  # source of flux
#         sink=M1.R1,  # target of flux
#         rate="300 mol/yr",  # weathering flux in
#         delta=d0,  # set a default flux
#     )

#     Connect(
#         source=M1.R1,  # source of flux
#         sink=M1.SI1,  # target of flux
#         rate="250 mol/yr",  # weathering flux in
#         delta=0,  # set a default flux
#     )

#     Connect(
#         source=M1.R1,  # source of flux
#         sink=M1.SI2,  # target of flux
#     )

#     M1.run(solver=solver)
# M1.get_delta_values()

#     assert M1.R1.c[-2] == c0
#     assert abs(M1.R1.d[-2] - d0) < 1e-10




def execute_e(model, lop, lor, lpc_f, lpc_r):
    """ """

    # numba.set_num_threads(2)

    # this has nothing todo with self.time below!

    dt: float = lor[0].mo.dt

    if model.first_start:
        start: float = process_time()
        (
            model.fn_vr,
            model.input_data,
            model.vr_data,
            model.vr_params,
            model.count,
        ) = build_vr_list(lpc_r)

        model.fn, model.da, model.pc = build_process_list(lor, lop)
        model.a, model.b, model.c, model.d, model.e = build_flux_lists_all(lor)
        """
        fn = typed list of functions (processes)
        da = process data
        pc = process parameters
        a = reservoir_list
        b = flux list
        c = direction list
        d = virtual reservoir list
        e = r0 list ???
        """
        model.first_start = False

        duration: float = process_time() - start
        print(f"\n Setup time {duration} cpu seconds\n")

    print("Starting solver")

    wts = time.time()
    start: float = process_time()

    if model.count > 0:
        if len(model.lop) > 0:
            foo(
                model.fn_vr,
                model.input_data,
                model.vr_data,
                model.vr_params,
                model.fn,
                model.da,
                model.pc,
                model.a,  # reservoir list
                model.b,  # flux list
                model.c,  # direction list
                model.d,  # r0 list
                model.e,
                model.time[:-1],
                model.dt,
            )
        else:
            foo_no_p(
                model.fn_vr,
                model.input_data,
                model.vr_data,
                model.vr_params,
                model.fn,
                model.a,  # reservoir list
                model.b,  # flux list
                model.c,  # direction list
                model.d,  # r0 list
                model.e,
                model.time[:-1],
                model.dt,
            )
    else:
        foo_no_vr(
            model.fn,
            model.da,
            model.pc,
            model.a,
            model.b,
            model.c,
            model.d,
            model.e,
            model.time[:-1],
            model.dt,
        )

    duration: float = process_time() - start
    wcd = time.time() - wts
    print(f"\n Total solver time {duration} cpu seconds, wt = {wcd}\n")



# from numba import jit
# @njit(parallel=False, fastmath=True, error_model="numpy")
def foo(fn_vr, input_data, vr_data, vr_params, fn, da, pc, a, b, c, d, e, maxt, dt):
    """
    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reservoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:
        # loop over function (process) list and compute fluxes
        j = 0
        for _ in enumerate(fn):
            fn[j](da[j], pc[j], i)
            j = j + 1

        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            # for j, r in enumerate(b):  # this will catch the list for each reservoir

            # sum fluxes in each reservoir
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        # # update reservoirs which do not depend on fluxes but on
        # # functions
        for j, f in enumerate(fn_vr):
            fn_vr[j](i, input_data[j], vr_data[j], vr_params[j])

        i = i + 1  # next time step


# @njit(parallel=False, fastmath=True, error_model="numpy")
def foo_no_p(fn_vr, input_data, vr_data, vr_params, fn, a, b, c, d, e, maxt, dt):
    """
    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reservoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:
        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            # for j, r in enumerate(b):  # this will catch the list for each reservoir

            # sum fluxes in each reservoir
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        # # update reservoirs which do not depend on fluxes but on
        # # functions
        for j, f in enumerate(fn_vr):
            fn_vr[j](i, input_data[j], vr_data[j], vr_params[j])

        i = i + 1  # next time step


# @njit(parallel=False, fastmath=True, error_model="numpy")
def foo_no_vr(fn, da, pc, a, b, c, d, e, maxt, dt):
    """Same as foo but no virtual reservoirs present

    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reservoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:
        # loop over processes
        j = 0
        for _ in enumerate(fn):
            fn[j](da[j], pc[j], i)
            j = j + 1

        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        i = i + 1


# @njit(parallel=False, fastmath=True, error_model="numpy")
def foo_no_vr_no_p(fn, a, b, c, d, e, maxt, dt):
    """Same as foo but no virtual reservoirs present

    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reservoir list
    e = r0 list ???
    """

    i = 1
    # calculate the resulting reservoir concentrations
    # summarize_fluxes(a, b, c, d, e, i, dt)
    r_steps: int = len(b)
    for _ in maxt:
        # loop over reservoirs
        for j in range(r_steps):
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        i = i + 1


def build_vr_list(lvr: list) -> tuple:
    """Build lists which contain all function references for
    virtual reservoirs as well aas their input values

    """

    fn = List()  # List() # list of functions
    input_data = List()  # reservoir data
    vr_data = List()  # flux data  flux.m flux.l, flux.h, flux.d
    vr_params = List()  # list of constants
    fn = numba.typed.List.empty_list(
        types.UniTuple(types.float64, 4)(
            types.int64,  # i
            types.ListType(types.float64[::1]),
            types.ListType(types.float64[::1]),
            types.ListType(types.float64),  # a3
        ).as_type()
    )

    count = 0
    for r in lvr:  # loop over reservoir processes
        func_name, in_d, vr_d, params = r.get_process_args()
        fn.append(func_name)
        input_data.append(in_d)
        vr_data.append(vr_d)
        vr_params.append(params)
        count = count + 1

    return fn, input_data, vr_data, vr_params, count


def build_flux_lists_all(lor, iso: bool = False) -> tuple:
    """flux_list :list [] contains all fluxes as
    [f.m, f.l, f.h, f.d], where each sublist relates to one reservoir

    i.e. per reservoir we have list [f1, f2, f3], where fi = [m, l]
    and m & l  = np.array()

    iso = False/True

    """

    r_list: list = List()
    v_list: list = List()
    r0_list: list = List()
    dir_list: list = List()
    rd_list: list = List()
    f_list: List = List()

    for r in lor:  # loop over all reservoirs
        if len(r.lof) > 0:
            rd_list = List([r.m, r.l, r.c, r.v])

            r_list.append(rd_list)
            v_list.append(float(r.volume))
            r0_list.append(float(r.sp.r))

            i = 0
            # add fluxes for each reservoir entry
            tf: list = List()  # temp list for flux data
            td: list = List()  # temp list for direction data

            # loop over all fluxes
            for f in r.lof:
                tf.append(f.fa)
                td.append(float(r.lodir[i]))
                i = i + 1

            f_list.append(tf)
            dir_list.append(td)

    return r_list, f_list, dir_list, v_list, r0_list


def build_process_list(lor: list, lop: list) -> tuple:
    from numba.typed import List
    import numba
    from numba.core import types

    fn = List()  # List() # list of functions
    da = List()  # data
    pc = List()  # list of constants

    print("Building Process List")

    tfn = numba.typed.List.empty_list(
        types.ListType(types.void)(  # return value
            types.ListType(types.float64[::1]),  # data array
            types.ListType(types.float64),  # parameter list
            types.int64,  # parameter 4
        ).as_type()
    )

    tda = List()  # temp list for data
    tpc = List()  # temp list for constants

    # note that types.List is differenfr from Types.ListType. Also
    # note that [::1]  declares C-style arrays see
    # https://numba.discourse.group/t/list-mistaken-as-list-when-creating-list-of-function-references/677/3

    for p in lop:  # loop over reservoir processes
        # print(f"working on {p.name}")
        func_name, data, proc_const = p.get_process_args()
        tfn.append(func_name)
        tda.append(data)
        tpc.append(proc_const)

    return tfn, tda, tpc


def apply_fractionation(
    f_m: str, c: Connection | Connect, icl: dict, ind3: str, ind2: str
) -> str:
    """If the connection involves isotope fractionation, add equation
    string that calculates the fractionation effect.

    :param f_m: string with the flux name
    :param c: connection object
    :param icl: dict of reservoirs that have actual fluxes
    :param ind2: indent 2 times
    :param ind3: indent 3 times

    :raises ValueError: if alpha is not between 0.9 and 1.1

    :returns eq: string with the fractionation equation

    Calculate the flux of the light isotope (f_l) as a function of the isotope
    ratios in the source reservoir soncentrations (s_c, s_l), and alpha (a) as

    f_l = s_l * f_m/ (a * s_c + s_l - a * s_l)
    """

    a = c.alpha / 1000 + 1
    s_c = get_ic(c.source, icl)
    s_l = get_ic(c.source, icl, isotopes=True)

    if int(c.alpha) != 0:
        eq = f"(\n{ind3} {s_l} * {f_m}\n{ind3} / ({a} * {s_c} + {s_l} - {a} * {s_l}))"
    else:
        eq = f"(\n{ind3} {f_m} * {s_l} / {s_c})"
    return eq


def apply_delta(
    f_m: str, c: Connection | Connect, icl: dict, ind3: str, ind2: str
) -> str:
    """If the connection involves a fixed deltae offset add equation
    string according to the delta value

    :param f_m: string with the flux name
    :param c: connection object
    :param icl: dict of reservoirs that have actual fluxes
    :param ind2: indent 2 times
    :param ind3: indent 3 times

    :raises ValueError: if alpha is not between 0.9 and 1.1

    :returns eq: string with the fractionation equation

    Calculate the flux of the light isotope (f_l) as a function of the isotope
    ratios in the source reservoir soncentrations (s_c, s_l), and alpha (a) as

    f_l = f_m * 1000/(r * (d + 1000) + 1000)
    """
    r = c.source.species.r
    d = c.delta
    exl = f"(\n" f"{ind3}{f_m} * 1000\n" f"{ind3} / ({r} * ({d} + 1000) + 1000))\n"
    return exl


# @njit(parallel=False, fastmath=True, error_model="numpy")
def calc_carbonates_1(i: int, input_data: List, vr_data: List, params: List) -> None:
    """Calculates and returns the carbonate concentrations and saturation state
     at the ith time-step of the model.

    The function assumes that vr_data will be in the following order:
        [H+, CA, HCO3, CO3, CO2(aq), omega]

    LIMITATIONS:
    - This in used in conjunction with ExternalCode objects!
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    See add_carbonate_system_1 in utility_functions.py on how to call this function

    Author: M. Niazi & T. Tsan, 2021
    """

    k1 = params[0]  # K1
    k2 = params[1]  # K2
    KW = params[2]  # KW
    KB = params[3]  # KB
    boron = params[4]  # boron
    ca2 = params[5]  # Ca2+
    ksp = params[6]  # Ksp

    dic: float = input_data[0][i - 1]
    ta: float = input_data[1][i - 1]
    hplus: float = vr_data[0][i - 1]

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1 * k1 - 4 * k1 * k2 * (1 - (2 * gamm))

    hplus: float = 0.5 * ((gamm - 1) * k1 + (dummy**0.5))
    # hco3 and co3
    """ Since CA = [hco3] + 2[co3], can the below expression can be simplified
    """
    # co3: float = dic / (1 + (hplus / k2) + ((hplus ** 2) / (k1 * k2)))
    hco3: float = dic / (1 + (hplus / k1) + (k2 / hplus))
    co3: float = (ca - hco3) / 2
    # co2 (aq)
    """DIC = hco3 + co3 + co2 + H2CO3 The last term is however rather
    small, so it may be ok to simply write co2aq = dic - hco3 + co3.
    Let's test this once we have a case where pco2 is calculated from co2aq
    """
    #  co2aq: float = dic / (1 + (k1 / hplus) + (k1 * k2 / (hplus ** 2)))
    co2aq: float = dic - hco3 - co3
    # omega: float = ca2 * co3 / ksp

    vr_data[0][i] = hplus
    vr_data[1][i] = ca
    vr_data[2][i] = hco3
    vr_data[3][i] = co3
    vr_data[4][i] = co2aq
    # vr_data[5][i] = oh
    # vr_data[5][i] = boh4
    # vr_data[5][i] = omega


# @njit(fastmath=True, error_model="numpy")
def calc_carbonates_2(i: int, input_data: List, vr_data: List, params: List) -> None:
    """Calculates and returns the carbonate concentrations and carbonate compensation
    depth (zcc) at the ith time-step of the model.

    The function assumes that vr_data will be in the following order:
        [H+, CA, HCO3, CO3, CO2(aq), zsat, zcc, zsnow, Fburial,
        B, BNS, BDS_under, BDS_resp, BDS, BCC, BPDC, BD,omega]

    LIMITATIONS:
    - This in used in conjunction with ExternalCode objects!
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    See add_carbonate_system_2 in utility_functions.py on how to call this function.
    The input data is a follows

        reservoir DIC.m,  # 0
        reservoir DIC.l,  # 1
        reservoir DIC.c,  # 2
        reservoir TA.m,  # 3 TA mass
        reservoir.TA.c,  # 4 TA concentration
        Export_flux.fa,  # 5
        area_table,  # 6
        area_dz_table,  # 7
        Csat_table,  # 8
        reservoir.DIC.v,  # 9 reservoir volume

    Author: M. Niazi & T. Tsan, 2021
    """

    # Parameters
    k1 = params[0]
    k2 = params[1]
    KW = params[2]
    KB = params[3]
    boron = params[4]
    ksp0 = params[5]
    kc = params[6]
    AD = params[8]
    zsat0 = int(abs(params[9]))
    ca2 = params[10]
    dt = params[11]
    I_caco3 = params[12]
    alpha = params[13]
    zsat_min = int(abs(params[14]))
    zmax = int(abs(params[15]))
    z0 = int(abs(params[16]))

    # Data
    dic: float = input_data[2][i - 1]  # DIC concentration [mol/kg]
    ta: float = input_data[4][i - 1]  # TA concentration [mol/kg]
    Bm: float = input_data[5][0]  # Carbonate Export Flux [mol/yr]
    B12: float = input_data[5][1]  # Carbonate Export Flux light isotope
    v: float = input_data[9][i - 1]  # volume
    # lookup tables
    depth_area_table: np.ndarray = input_data[6]  # depth look-up table
    area_dz_table: np.ndarray = input_data[7]  # area_dz table
    Csat_table: np.ndarray = input_data[8]  # Csat table

    # vr_data
    hplus: float = vr_data[0][i - 1]  # H+ concentration [mol/kg]
    zsnow = vr_data[7][i - 1]  # previous zsnow

    # calc carbonate alkalinity based t-1
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # calculate carbon speciation
    # The following equations are after Follows et al. 2006
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1 * k1 - 4 * k1 * k2 * (1 - (2 * gamm))

    hplus: float = 0.5 * ((gamm - 1) * k1 + (dummy**0.5))
    # co3: float = dic / (1 + (hplus / k2) + ((hplus ** 2) / (k1 * k2)))
    hco3: float = dic / (1 + (hplus / k1) + (k2 / hplus))
    co3: float = (ca - hco3) / 2
    # DIC = hco3 + co3 + co2 + H2CO3 The last term is however rather
    # small, so it may be ok to simply write co2aq = dic - hco3 + co3.
    co2aq: float = dic - co3 - hco3
    # co2aq: float = dic / (1 + (k1 / hplus) + (k1 * k2 / (hplus ** 2)))
    # omega: float = (ca2 * co3) / ksp

    # ---------- compute critical depth intervals eq after  Boudreau (2010)
    # all depths will be positive to facilitate the use of lookup_tables

    # prevent co3 from becoming zero
    if co3 <= 0:
        co3 = 1e-16

    zsat = int(max((zsat0 * np.log(ca2 * co3 / ksp0)), zsat_min))  # eq2
    if zsat < zsat_min:
        zsat = int(zsat_min)

    zcc = int(zsat0 * np.log(Bm * ca2 / (ksp0 * AD * kc) + ca2 * co3 / ksp0))  # eq3

    # ---- Get fractional areas
    B_AD = Bm / AD

    if zcc > zmax:
        zcc = int(zmax)
    if zcc < zsat_min:
        zcc = zsat_min

    A_z0_zsat = depth_area_table[z0] - depth_area_table[zsat]
    A_zsat_zcc = depth_area_table[zsat] - depth_area_table[zcc]
    A_zcc_zmax = depth_area_table[zcc] - depth_area_table[zmax]

    # ------------------------Calculate Burial Fluxes------------------------------------
    # BCC = (A(zcc, zmax) / AD) * B, eq 7
    BCC = A_zcc_zmax * B_AD

    # BNS = alpha_RD * ((A(z0, zsat) * B) / AD) eq 8
    BNS = alpha * A_z0_zsat * B_AD

    # BDS_under = kc int(zcc,zsat) area' Csat(z,t) - [CO3](t) dz, eq 9a
    diff_co3 = Csat_table[zsat:zcc] - co3
    area_p = area_dz_table[zsat:zcc]

    BDS_under = kc * area_p.dot(diff_co3)
    BDS_resp = alpha * (A_zsat_zcc * B_AD - BDS_under)
    BDS = BDS_under + BDS_resp

    # BPDC =  kc int(zsnow,zcc) area' Csat(z,t) - [CO3](t) dz, eq 10
    if zcc < zsnow:  # zcc cannot
        if zsnow > zmax:  # zsnow cannot exceed ocean depth
            zsnow = zmax

        diff = Csat_table[zcc : int(zsnow)] - co3
        area_p = area_dz_table[zcc : int(zsnow)]

        BPDC = kc * area_p.dot(diff)
        zold = BPDC / (area_dz_table[int(zsnow)] * I_caco3) * dt
        # eq 4 dzsnow/dt = Bpdc(t) / (a'(zsnow(t)) * ICaCO3
        zsnow = zsnow - BPDC / (area_dz_table[int(zsnow)] * I_caco3) * dt

    else:  # zcc > zsnow
        # there is no carbonate below zsnow, so BPDC = 0
        zsnow = zcc
        BPDC = 0

    # BD & F_burial
    BD: float = BDS + BCC + BNS + BPDC
    Fburial = Bm - BD
    Fburial_l = Fburial * input_data[1][i - 1] / input_data[0][i - 1]
    diss = (Bm - Fburial) * dt  # dissolution flux
    diss12 = (B12 - Fburial_l) * dt  # dissolution flux light isotope

    # # print("{Fburial}.format(")
    # print(Bm)
    # print(Fburial)
    # print(diss)
    # print()
    # # print('df ={:.2e}\n'.format(diss/dt))

    """ Now that the fluxes are known we need to update the reservoirs.
    The concentration in the in the DIC (and TA) of this box are
    DIC.m[i] + Export Flux - Burial Flux, where the isotope ratio
    the Export flux is determined by the overlying box, and the isotope ratio
    of the burial flux is determined by the isotope ratio of this box
    """

    # Update DIC in the deep box
    input_data[0][i] = input_data[0][i] + diss  # DIC
    input_data[1][i] = input_data[1][i] + diss12  # 12C
    input_data[2][i] = input_data[0][i] / v  # DIC concentration

    # Update TA in deep box
    input_data[3][i] = input_data[3][i] + 2 * diss  # TA
    input_data[4][i] = input_data[3][i] / v  # TA concentration

    # copy results into datafields
    vr_data[0][i] = hplus  # 0
    vr_data[1][i] = ca  # 1
    vr_data[2][i] = hco3  # 2
    vr_data[3][i] = co3  # 3
    vr_data[4][i] = co2aq  # 4
    vr_data[5][i] = zsat  # 5
    vr_data[6][i] = zcc  # 6
    vr_data[7][i] = zsnow  # 7
    vr_data[8][i] = Fburial  # 8
    vr_data[9][i] = Fburial_l  # 9
    vr_data[10][i] = diss / dt  # 9
    vr_data[11][i] = Bm  # 9



"""esbmtk: A general purpose Earth Science box model toolkit Copyright
     (C), 2020-2021 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
     published by the Free Software Foundation, either version 3 of
     the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see
     <https://www.gnu.org/licenses/>.

"""
from __future__ import annotations
import typing as tp
import numpy as np
from .esbmtk_base import esbmtkBase

if tp.TYPE_CHECKING:
    from .esbmtk import Reservoir, Model
    from .extended_classes import ReservoirGroup


# define a transform function to display the Hplus concentration as pH
def phc(m: float) -> float:
    """the reservoir class accepts a plot transform. here we use this to
    display the H+ concentrations as pH. After import, you can use it
    with like this in the reservoir definition

     plot_transform_c=phc,

    """
    import numpy as np

    pH = -np.log10(m)
    return pH


class SeawaterConstants(esbmtkBase):
    """Provide basic seawater properties as a function of T, P and Salinity.
    This module does not reca

    Example:

    Seawater(name="SW",
             register=M # model handle
             temperature = optional in C, defaults to 25,
             salinity  = optional in psu, defaults to 35,
             pressure = optional, defaults to 0 bars = 1atm,
             pH = optional, defaults to 8.1,
            )

    Results are always in mol/kg

    Acess the values "dic", "ta", "ca", "co2", "hco3",
    "co3", "boron", "boh4", "boh3", "oh", "ca2", "so4","hplus",
    as SW.co3 etc.

    This method also provides "K0", "K1", "K2", "KW", "KB", "Ksp",
    "Ksp0", "KS", "KF" and their corresponding pK values, as well
    as the density for the given (P/T/S conditions)

    useful methods:

    SW.show() will list all known values

    After initialization this class provides access to each value the following way

    instance_name.variable_name

    """

    def __init__(self, **kwargs: dict[str, str]):

        from esbmtk import Model, Reservoir, ReservoirGroup

        self.defaults: dict[list[any, tuple]] = {
            "name": ["None", (str)],
            "salinity": [35.0, (int, float)],
            "temperature": [25.0, (int, float)],
            "pH": [8.1, (int, float)],
            "pressure": [0, (int, float)],
            "register": ["None", (Model, Reservoir, ReservoirGroup)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["name"]
        self.__initialize_keyword_variables__(kwargs)
        self.parent = self.register

        # legacy names
        self.n: str = self.name  # string =  name of this instance
        # self.mo: Model = self.model
        self.hplus = 10**-self.pH
        self.constants: list = ["K0", "K1", "K2", "KW", "KB", "Ksp", "Ksp0", "KS", "KF"]
        self.species: list = [
            "dic",
            "ta",
            "ca",
            "co2",
            "co2aq",
            "hco3",
            "co3",
            "boron",
            "boh4",
            "boh3",
            "oh",
            "ca2",
            "so4",
            "hplus",
        ]

        self.update_parameters()

        if self.register != "None":
            self.__register_name_new__()

    def update_parameters(self, **kwargs: dict) -> None:
        """Update values if necessary"""

        if kwargs:
            self.__initialize_keyword_variables__(kwargs)

        # update K values and species concentrations according to P, S, and T
        self.__get_density__()
        self.__init_std_seawater__()
        self.__init_bisulfide__()
        self.__init_hydrogen_floride__()
        self.__init_carbon__()
        self.__init_boron__()
        self.__init_water__()
        self.__init_gasexchange__()
        self.__init_calcite__()
        self.__init_c_fractionation_factors__()

        # get total alkalinity
        self.ca = self.hco3 + 2 * self.co3
        self.ta = self.ca + self.boh4 + self.oh - self.hplus

    def show(self) -> None:
        """Printout constants"""

        from math import log10

        for n in self.species:
            v = getattr(self, n)
            print(f"{n} = {v * 1E6:.2f} nmol/kg")

        print()
        print(f"pCO2 = {get_pco2(self):.2e}")
        print(f"pH = {-log10(self.hplus):.2f}")
        print(f"salinity = {self.salinity:.2f}")
        print(f"temperature = {self.temperature:.2f}\n")

        for n in self.constants:
            K = getattr(self, n)  # get K value
            pk = f"p{n.lower()}"  # get K name
            print(f"{n} = {K:.2e}, {pk} = {-log10(K):.2f}")

    def __get_density__(self):
        """Calculate seawater density as function of temperature,
        pressure and salinity in kg/m^3. Shamelessy copied
        from R. Zeebes equic.m mathlab file.

        TC = temp in C
        P = pressure
        S = salinity
        """

        TC = self.temperature
        P = self.pressure
        S = self.salinity
        # density of pure water
        rhow = (
            999.842594
            + 6.793952e-2 * TC
            - 9.095290e-3 * TC**2
            + 1.001685e-4 * TC**3
            - 1.120083e-6 * TC**4
            + 6.536332e-9 * TC**5
        )

        # density of of seawater at 1 atm, P=0
        A = (
            8.24493e-1
            - 4.0899e-3 * TC
            + 7.6438e-5 * TC**2
            - 8.2467e-7 * TC**3
            + 5.3875e-9 * TC**4
        )
        B = -5.72466e-3 + 1.0227e-4 * TC - 1.6546e-6 * TC**2
        C = 4.8314e-4
        rho0 = rhow + A * S + B * S ** (3 / 2) + C * S**2

        """Secant bulk modulus of pure water is the average change in
        pressure divided by the total change in volume per unit of
        initial volume.
        """
        Ksbmw = (
            19652.21
            + 148.4206 * TC
            - 2.327105 * TC**2
            + 1.360477e-2 * TC**3
            - 5.155288e-5 * TC**4
        )
        # Secant bulk modulus of seawater at 1 atm
        Ksbm0 = (
            Ksbmw
            + S * (54.6746 - 0.603459 * TC + 1.09987e-2 * TC**2 - 6.1670e-5 * TC**3)
            + S ** (3 / 2) * (7.944e-2 + 1.6483e-2 * TC - 5.3009e-4 * TC**2)
        )
        # Secant modulus of seawater at S,T,P
        Ksbm = (
            Ksbm0
            + P
            * (3.239908 + 1.43713e-3 * TC + 1.16092e-4 * TC**2 - 5.77905e-7 * TC**3)
            + P * S * (2.2838e-3 - 1.0981e-5 * TC - 1.6078e-6 * TC**2)
            + P * S ** (3 / 2) * 1.91075e-4
            + P * P * (8.50935e-5 - 6.12293e-6 * TC + 5.2787e-8 * TC**2)
            + P**2 * S * (-9.9348e-7 + 2.0816e-8 * TC + 9.1697e-10 * TC**2)
        )
        # Density of seawater at S,T,P in kg/m^3
        self.density = rho0 / (1.0 - P / Ksbm)

    def __init_std_seawater__(self) -> None:
        """Provide values for standard seawater. Data after Zeebe and Gladrow
        all values in mol/kg. All values after Zeebe and Gladrow 2001

        """

        self.dic = 0.00204
        self.boron = 0.00042
        self.oh = 0.00001
        self.so4 = 2.7123 / 96
        self.ca2 = 0.01028
        self.Ksp0 = 4.29e-07  # after after Boudreau et al 2010

    def __init_hydrogen_floride__(self) -> None:
        """Bisulfide ion concentration after Dickson 1994, cf.
        Zeebe and Gladrow 2001, p 260

        """

        import numpy as np

        T = 273.15 + self.temperature
        S = self.salinity
        I = (19.924 * S) / (1000 - 1.005 * S)

        lnKF = (
            1590.2 / T
            - 12.641
            + 1.525 * I**0.5
            + np.log(1 - 0.001005 * S)
            + np.log(1 + self.ST / self.KS)
        )

        self.KF = np.exp(lnKF)
        self.FT = 7e-5 * self.salinity / 35

    def __init_bisulfide__(self) -> None:
        """Bisulfide ion concentration after Dickson 1994, cf.
        Zeebe and Gladrow 2001, p 260

        """

        import numpy as np

        T = 273.15 + self.temperature
        S = self.salinity
        I = (19.924 * S) / (1000 - 1.005 * S)
        lnKS = (
            -4276.1 / T
            + 141.328
            - 23.093 * np.log(T)
            + (-13856 / T + 324.57 - 47.986 * np.log(T)) * I**0.5
            + (35474 / T - 771.54 + 114.723 * np.log(T)) * I
            - 2698 / T * I**1.5
            + 1776 / T * I**2
            + np.log(1 - 0.001005 * S)
        )

        self.KS = np.exp(lnKS)
        self.ST = self.so4 * self.salinity / 35

    def __init_gasexchange__(self) -> None:
        """Initialize constants for gas-exchange processes"""

        self.water_vapor_partial_pressure()
        self.co2_solubility_constant()
        self.o2_solubility_constant()

    def __init_carbon__(self) -> None:
        """Calculate the carbon equilibrium values as function of
        temperature T and salinity S

        """

        from math import exp, log

        T = 273.15 + self.temperature
        S = self.salinity

        # After Weiss 1974
        lnK0: float = (
            93.4517 * 100 / T
            - 60.2409
            + 23.3585 * log(T / 100)
            + S * (0.023517 - 0.023656 * T / 100 + 0.0047036 * (T / 100) ** 2)
        )

        lnk1: float = (
            -2307.1266 / T
            + 2.83655
            - 1.5529413 * log(T)
            + S**0.5 * (-4.0484 / T - 0.20760841)
            + S * 0.08468345
            + S ** (3 / 2) * -0.00654208
            + log(1 - 0.001006 * S)
        )

        lnk2: float = (
            -9.226508
            - 3351.6106 / T
            - 0.2005743 * log(T)
            + (-0.106901773 - 23.9722 / T) * S**0.5
            + 0.1130822 * S
            - 0.00846934 * S**1.5
            + log(1 - 0.001006 * S)
        )

        self.K0: float = exp(lnK0)
        self.K1: float = exp(lnk1)
        self.K2: float = exp(lnk2)

        self.K1 = self.__pressure_correction__("K1", self.K1)
        self.K2 = self.__pressure_correction__("K2", self.K2)

        self.K1K1: float = self.K1 * self.K1
        self.K1K2: float = self.K1 * self.K2

        # self.K_l : list = [self.K0, self.K1, self.K2, self.K1K1, self.K1K2]

        self.co2 = self.dic / (
            1 + self.K1 / self.hplus + self.K1 * self.K2 / self.hplus**2
        )
        self.hco3 = self.dic / (1 + self.hplus / self.K1 + self.K2 / self.hplus)
        self.co3 = self.dic / (
            1 + self.hplus / self.K2 + self.hplus**2 / (self.K1 * self.K2)
        )
        self.co2aq = self.dic / (
            1 + (self.K1 / self.hplus) + (self.K1 * self.K2 / (self.hplus**2))
        )

    def __init_boron__(self) -> None:
        """Calculate the boron equilibrium values as function of
        temperature T and salinity S

        """

        from math import exp, log

        T = 273.15 + self.temperature
        S = self.salinity

        lnkb = (
            (
                -8966.9
                - 2890.53 * S**0.5
                - 77.942 * S
                + 1.728 * S**1.5
                - 0.0996 * S**2
            )
            / T
            + 148.0248
            + 137.1942 * S**0.5
            + 1.62142 * S
            - (24.4344 + 25.085 * S**0.5 + 0.2474 * S) * log(T)
            + 0.053105 * S**0.5 * T
        )

        self.KB = exp(lnkb)
        self.KB = self.__pressure_correction__("KB", self.KB)

        self.boh4 = self.boron * self.KB / (self.hplus + self.KB)
        self.boh3 = self.boron - self.boh4

    def __init_water__(self) -> None:
        """Calculate the water equilibrium values as function of
        temperature T and salinity S

        """

        from math import exp, log

        T = 273.15 + self.temperature
        S = self.salinity

        lnKW = (
            148.96502
            - 13847.27 / T
            - 23.6521 * log(T)
            + (118.67 / T - 5.977 + 1.0495 * log(T)) * S**0.5
            - 0.01615 * S
        )
        self.KW = exp(lnKW)
        self.KW = self.__pressure_correction__("KW", self.KW)
        self.oh = self.KW / self.hplus

    def __pressure_correction__(self, n: str, K: float) -> float:
        """Correct K-values for pressure. After Zeebe and Wolf Gladrow 2001

        name = name of K-value, i.e. "K1"
        K = uncorrected value
        T = temperature in Deg C
        P = pressure in atm
        """

        from math import exp, log

        R: float = 83.131
        Tc: float = self.temperature
        T: float = 273.15 + Tc
        P: float = self.pressure
        RT: float = R * T

        A: dict = {}
        A["K1"]: list = [25.50, 0.1271, 0.0, 3.08, 0.0877]
        A["K2"]: list = [15.82, -0.0219, 0.0, -1.13, -0.1475]
        A["KB"]: list = [29.48, 0.1622, -2.6080, 2.84, 0.0]
        A["KW"]: list = [25.60, 0.2324, -3.6246, 5.13, 0.0794]
        A["KS"]: list = [18.03, 0.0466, 0.3160, 4.53, 0.0900]
        A["KF"]: list = [9.780, -0.0090, -0.942, 3.91, 0.054]
        A["Kca"]: list = [48.76, 0.5304, 0.0, 11.76, 0.3692]
        A["Kar"]: list = [46.00, 0.5304, 0.0, 11.76, 0.3692]

        a: list = A[n]

        DV: float = -a[0] + (a[1] * Tc) + (a[2] / 1000 * Tc**2)
        DK: float = -a[3] / 1000 + (a[4] / 1000 * Tc) + (0 * Tc**2)

        # print(f"DV = {DV}")
        # print(f"DK = {DK}")
        # print(f"log k= {log(K)}")

        lnkp: float = -(DV / RT) * P + (0.5 * DK / RT) * P**2 + log(K)
        # print(lnkp)

        return exp(lnkp)

    def water_vapor_partial_pressure(self) -> None:
        """Calculate the water vapor partial pressure at sealevel (1 atm) as
        a function of temperature and salinity. Eq. Weiss and Price 1980
        doi:10.1016/0304-4203(80)90024-9

        Since we assume that we only use this expression at sealevel,
        we drop the pressure term

        The result is in p/1atm (i.e., a percentage)
        """

        T = self.temperature + 273.15
        S = self.salinity

        self.p_H2O = np.exp(
            24.4543 - 67.4509 * (100 / T) - 4.8489 * np.log(T / 100) - 0.000544 * S
        )

    def co2_solubility_constant(self) -> None:
        """Calculate the solubility of CO2 at a given temperature and salinity.
        Coefficients after Sarmiento and Gruber 2006 which includes
        corrections for CO2 to correct for non ideal gas behavior

        Parameters Ai & Bi from Tab 3.2.2 in  Sarmiento and Gruber 2006

        The result is in mol/(m^3 * atm)
        """

        # Calculate the volumetric solubility function F_A in mol/l
        S = self.salinity  # unitless
        T = 273.15 + self.temperature  # C
        A1 = -160.7333
        A2 = 215.4152
        A3 = 89.892
        A4 = -1.47759
        B1 = 0.029941
        B2 = -0.027455
        B3 = 0.0053407

        # F in mol/(l * atm)
        F = self.calc_solubility_term(S, T, A1, A2, A3, A4, B1, B2, B3)

        # correct for water vapor partial pressure
        self.SA_co2 = F / (1 - self.p_H2O)  # mol/(m^3 * atm)

    def o2_solubility_constant(self) -> None:
        """Calculate the solubility of CO2 at a given temperature and salinity. Coefficients
        after Sarmiento and Gruber 2006 which includes corrections for CO2 to correct for non
        ideal gas behavior

        Parameters Ai & Bi from Tab 3.2.2 in  Sarmiento and Gruber 2006

        The result is in mol/(m^3 atm)
        """

        # Calculate the volumetric solubility function F_A in mol/l/m^3
        S = self.salinity  # unit less
        T = 273.15 + self.temperature  # in C
        A1 = -58.3877
        A2 = 85.8079
        A3 = 23.8439
        A4 = 0
        B1 = -0.034892
        B2 = 0.015568
        B3 = -0.0019387

        b = self.calc_solubility_term(S, T, A1, A2, A3, A4, B1, B2, B3)

        # and convert from bunsen coefficient to solubility
        VA = 22.4136  # after Sarmiento & Gruber 2006
        self.SA_o2 = b / VA

    def calc_solubility_term(self, S, T, A1, A2, A3, A4, B1, B2, B3) -> float:
        ln_F = (
            A1
            + A2 * (100 / T)
            + A3 * np.log(T / 100)
            + A4 * (T / 100) ** 2
            + S * (B1 + B2 * (T / 100) + B3 * (T / 100) ** 2)
        )
        F = np.exp(ln_F) * 1000  # to get mol/(m^3 atm)

        return F

    def __init_calcite__(self) -> None:
        """Calculate Calcite solubility as a function of pressure following
        Fig 1 in in Boudreau et al, 2010, https://doi.org/10.1029/2009gl041847

        Note that this equation assumes an idealized ocean temperature profile.
        So it cannot be applied to a warm ocean

        """

        self.Ksp = 4.3513e-7 * np.exp(0.0019585 * self.pressure)

    def __init_c_fractionation_factors__(self):
        """Calculate the fractionation factors for the various carbon species transitions.
        After Zeebe and Gladrow, 2001, Chapter 3.2.3, page 186

        e = (a -1) * 1E3

        and

        a =  1 + e / 1E3

        where the subscripts denote:

        g = gaseous CO2
        d = dissolved CO2
        b = bicarbonate ion
        c = carbonate ion

        """

        T = 273.15 + self.temperature

        # CO2g versus HCO3, e = epsilon, a = alpha
        self.e_gb: float = -9483 / T + 23.89
        self.a_gb: float = 1 + self.e_gb / 1000

        # CO2aq versus CO2g
        self.e_dg: float = -373 / T + 0.19
        self.a_dg: float = 1 + self.e_dg / 1000

        # CO2aq versus HCO3
        self.e_db: float = -9866 / T + 24.12
        self.a_db: float = 1 + self.e_db / 1000

        # CO32- versus HCO3
        self.e_cb: float = -867 / T + 2.52
        self.a_cb: float = 1 + self.e_cb / 1000

        # kinetic fractionation during gas exchange
        # parameters after Zhang et. al.1995
        # https://doi.org/10.1016/0016-7037(95)91550-D
        m = 0.14 / 16
        c = m * 5 + 0.95
        self.e_u: float = self.temperature * m - c
        self.a_u: float = 1 + self.e_u / 1000


"""
    dic: tp.Union[Reservoir, esbmtk.extended_classes.VirtualReservoir],
    hplus: tp.Union[Reservoir, esbmtk.extended_classes.VirtualReservoir],
    SW: SeawaterConstants,

"""


def get_pco2(SW) -> float:
    """Calculate the concentration of pCO2"""

    dic_c: float = SW.dic
    hplus_c: float = SW.hplus
    k1: float = SW.K1
    k2: float = SW.K2
    co2: np.ndarray = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))
    pco2: np.ndarray = co2 / SW.K0 * 1e6
    return pco2


def calc_pCO2(
    dic,  # see above why no type hints
    hplus,
    SW,
) -> np.ndarray:

    """
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic: Reservoir  = DIC concentrations in mol/kg
    hplus: Reservoir = H+ concentrations in mol/kg
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2(dic,h,SW)

     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )
    Author: T. Tsan

    """

    dic_c: np.ndarray = dic.c
    hplus_c: np.ndarray = hplus.c

    k1: float = SW.K1
    k2: float = SW.K2

    co2: np.ndarray = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))

    pco2: np.ndarray = co2 / SW.K0 * 1e6

    return pco2


def calc_pCO2b(
    dic: np.ndarray,
    hplus: np.ndarray,
    SW: SeawaterConstants,
) -> np.ndarray:

    """
    Same as calc_pCO2, but accepts values/arrays rather than Reservoirs.
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic:  = DIC concentrations in mol/kg
    hplus: = H+ concentrations in mol/kg
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2b(dic,h,SW)
     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2b,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )
    """

    dic_c: np.ndarray = dic

    hplus_c: np.ndarray = hplus

    k1: float = SW.K1
    k2: float = SW.K2

    co2: np.ndarray = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))

    pco2: np.ndarray = co2 / SW.K0 * 1e6

    return pco2
